from .base_navigator import BaseNavigator
from .book_navigator import BookNavigator, BookNavigatorException
from .clip_navigator import ClipNavigator
from .section_navigator import SectionNavigator
from .toc_navigator import TocNavigator

__all__ = ["BaseNavigator", "BookNavigator", "BookNavigatorException", "ClipNavigator", "SectionNavigator", "TocNavigator"]



from dataclasses import dataclass, field
from typing import List

from loguru import logger

from utilities.domlib import Document
from models.audio import Audio
from models.reference import Reference
from models.section import Section
from models.text import Text
from sources.source import DtbSource


@dataclass
class Smil:
    """This class represents a SMIL file."""

    source: DtbSource
    reference: Reference

    # Internal attributes (dynamically populated)
    _title: str = field(init=False, default="")
    _total_duration: float = field(init=False, default=0.0)
    _is_parsed: bool = field(init=False, default=False)
    _sections: List[Section] = field(init=False, default_factory=list)

    def __post_init__(self): ...

    @property
    def title(self) -> str:
        if not self._is_parsed:
            self._parse()
        return self._title

    @property
    def total_duration(self) -> str:
        if not self._is_parsed:
            self._parse()
        return self._total_duration

    @property
    def sections(self) -> List[Section]:
        if not self._is_parsed:
            self._parse()
        return self._sections

    def get_full_text(self) -> str:
        result = []
        if self._is_parsed is False:
            self._parse()
        for par in self._sections:
            result.append(par.text._parse())

        return "\n".join(result)

    def _parse(self) -> None:
        """Load a the SMIL file (if not already loaded) and parse it."""
        if self._is_parsed:
            logger.debug(f"SMIL '{self.reference.resource}' is already loaded.")
            return

        # Get the resource data
        data = self.source.get(self.reference.resource)

        if data is None:
            logger.debug(f"Could not get SMIL '{self.reference.resource}'.")
            return

        if not isinstance(data, Document):
            logger.debug(f"No Document to process ({self.reference.resource}).")
            return

        # Title
        elt = data.get_elements_by_tag_name("meta", {"name": "dc:title"}).first()
        if elt:
            self._title = elt.get_attr("content")
            logger.debug(f"SMIL '{self.reference.resource}' title set : '{self._title}'.")

        # Total duration
        elt = data.get_elements_by_tag_name("meta", {"name": "ncc:timeInThisSmil"}).first()
        if elt:
            duration = elt.get_attr("content")
            h, m, s = duration.split(":")
            self._total_duration = float(h) * 3600 + float(m) * 60 + float(s)
            logger.debug(f"SMIL {self.reference.resource} duration set : {self._total_duration}s.")

        # Process sequences in body
        for body_seq in data.get_elements_by_tag_name("seq", having_parent_tag_name="body").all():
            # Process the <par/> element in the sequence
            for par in body_seq.get_children_by_tag_name("par").all():
                par_id = par.get_attr("id")

                # Handle the <text/>
                text = par.get_children_by_tag_name("text").first()
                id = text.get_attr("id")
                reference = Reference.create_href_or_src(text.get_attr("src"))
                current_text = Text(self.source, id, reference)
                current_par = Section(self.source, par_id, current_text)

                # Handle the <audio/> clip
                for par_seq in par.get_children_by_tag_name("seq").all():
                    for audio in par_seq.get_children_by_tag_name("audio").all():
                        id = audio.get_attr("id")
                        src = audio.get_attr("src")
                        begin = float(audio.get_attr("clip-begin")[4:-1])
                        end = float(audio.get_attr("clip-end")[4:-1])
                        current_par._clips.append(Audio(self.source, id, src, begin, end))
                    logger.debug(f"SMIL {self.reference.resource}, par: {current_par.id} contains {len(current_par._clips)} clip(s).")

                # Add to the list of Parallel
                self._sections.append(current_par)

        self._is_parsed = True
        logger.debug(f"SMIL {self.reference.resource} contains {len(self._sections)} pars.")
        logger.debug(f"SMIL {self.reference.resource} sucessfully loaded.")



import sys
from enum import StrEnum

from loguru import logger


class LogLevel(StrEnum):
    """A simple class to handle the logging level."""

    NONE = ""
    TRACE = "TRACE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    SUCCESS = "SUCCESS"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"

    @staticmethod
    def set(level: "LogLevel") -> None:
        """Set the logging level.

        Args:
            level (LogLevel): the wanted logging level.
        """
        logger.remove()
        if level != LogLevel.NONE:
            logger.add(sys.stderr, level=level.value)



from dataclasses import dataclass
from typing import Union


@dataclass
class Reference:
    """This class represents a reference to a fragment in a file."""

    resource: str
    fragment: str

    @staticmethod
    def create_href_or_src(string: str) -> Union["Reference", None]:
        """Create a Reference from a string.

        Example:
        - "dijn0159.smil#mxhp_0001" will return Reference(resource="dijn0159.smil", fragment="mxhp_0001").


        Returns:
            Union[Reference, None]: the Reference.
        """
        if "#" not in string:
            return None
        source, fragment = string.split("#")
        return Reference(source, fragment)



"""Cache statistics"""

from dataclasses import dataclass, field
from typing import List, cast


@dataclass
class _CacheStatItem:
    """This classe represents cache statistics item.

    It is intended for internal use.
    """

    name: str = field(init=True)
    hits: int = field(init=False, default=0)
    queries: int = field(init=False, default=1)

    @property
    def efficiency(self) -> float:
        return self.hits / self.queries


@dataclass
class CacheStats:
    # Private attributs
    _items: List[_CacheStatItem] = field(init=False, default_factory=list)

    def hit(self, resource_name: str) -> None:
        item = _CacheStatItem(resource_name)
        item.hits = 1
        self._add(item)

    def miss(self, resource_name: str) -> None:
        item = _CacheStatItem(resource_name)
        item.hits = 0
        self._add(item)

    def _add(self, item: _CacheStatItem) -> None:
        """Add a statistics item to the cache

        Args:
            item (CacheStatItem): the item to add
        """
        try:
            item_index = [_.name for _ in self._items].index(item.name)
            cache_stat_item = self._items[item_index]
            cache_stat_item.queries += 1
            cache_stat_item.hits = cache_stat_item.hits + item.hits
        except ValueError:
            self._items.append(item)

    def get_stats(self) -> dict:
        """Get the cache statistics.

        Returns:
            dict: a dictionary holding the global stats and the details
        """
        hit_count = sum([_.hits for _ in self._items])
        query_count = sum([_.queries for _ in self._items])
        result = {
            "cached_items": len(self._items),
            "total_queries": query_count,
            "total_hits": hit_count,
            "cache_efficiency": hit_count / query_count if query_count else 0,
            "details": [],
        }

        self._items.sort(key=lambda x: x.name)
        for item in self._items:
            detail = {
                "item_name": item.name,
                "queries": item.queries,
                "hits": item.hits,
                "efficiency": item.efficiency,
            }
            cast(list, result["details"]).append(detail)
        return result



from typing import Union
from utilities.domlib import Document
from utilities.fetcher import Fetcher
from sources.source import DtbSource


class FolderDtbSource(DtbSource):
    """This class gets data from a filesystem folder or a web location"""

    def __init__(self, base_path: str, initial_cache_size=0) -> None:
        base_path = base_path if base_path.endswith("/") else f"{base_path}/"
        super().__init__(base_path, initial_cache_size)

        if Fetcher.is_available(base_path) is False:
            raise FileNotFoundError

    def get(self, resource_name: str) -> Union[bytes, Document, None]:
        path = f"{self._base_path}{resource_name}"

        # Try to get data from the cached resources
        cached_data = self._cache.get(resource_name)
        if cached_data is not None:
            return cached_data

        data = Fetcher.fetch(path)

        # Try to create a Document
        doc = DtbSource.convert_to_document(data)

        # Eventualy cache the resource
        self.do_cache(resource_name, doc)

        return doc



from typing import List
from models import Section
from navigators.base_navigator import BaseNavigator


class SectionNavigator(BaseNavigator):
    def first(self) -> Section:
        return super().first()

    def last(self) -> Section:
        return super().last()

    def next(self) -> Section:
        return super().next()

    def prev(self) -> Section:
        return super().prev()

    def current(self) -> Section:
        return super().current()

    def all(self) -> List[Section]:
        return super().all()

    def navigate_to(self, item_id) -> Section:
        return super().current()



"""Daisy Book related classes"""

from dataclasses import dataclass, field
from typing import List, Union

from loguru import logger

from utilities.domlib import Document

from models import MetaData, Reference, Smil, TocEntry
from sources import DtbSource


class DaisyBookException(Exception):
    """Daisy Book Exception class."""

    def __init__(self, message):
        super().__init__(message)


@dataclass
class DaisyBook:
    """Representation of a Daisy 2.02 Digital Talking Book file.

    Raises:
        DaisyBookException: this exception is raised when the instance cannot be set up.
    """

    source: DtbSource

    # Internal attributes
    _title: str = field(init=False, default="")
    _navigation_depth: int = field(init=False, default=0)
    _metadata: List[MetaData] = field(init=False, default_factory=list)
    _toc_entries: List[TocEntry] = field(init=False, default_factory=list)
    _smils: List[Smil] = field(init=False, default_factory=list)

    def __post_init__(self):
        """DaisyBook instance post-initialization.

        The steps are :
        - Load the NCC.html file, then
            - Extract the TOC entries
            - Extact the metadata
            - Create the SMIL entries
            - Set the books title and the navigation depth

        Raises:
            DaisyBookException: this exception is raised when the instance cannot be set up.
        """
        # Get the ncc.html file content
        ncc_document = self.source.get("ncc.html")

        # No data, no further processing !
        if ncc_document is None or not isinstance(ncc_document, Document):
            message = f"Could not process {self.source.base_path}."
            logger.critical(message)
            raise DaisyBookException(f"Could not process {message}.")

        # Populate the entries list
        self._populate_entries(ncc_document)

        # Populate the metadata list
        self._populate_metadata(ncc_document)

        # Populate the smils list
        for entry in self._toc_entries:
            self._smils.append(entry.smil)

        # Set the title and navigation depth from the metadata
        for meta in self._metadata:
            match meta.name:
                case "dc:title":
                    self._title = meta.content
                case "ncc:depth":
                    self._navigation_depth = int(meta.content)

    @property
    def cache_stats(self) -> dict:
        """Get the cache statistics.

        Returns:
            dict: the cache statistics.
        """
        return self.source._cache.get_stats()

    @property
    def title(self) -> str:
        """Get the books title.

        Returns:
            str: the books title.
        """
        return self._title

    @property
    def navigation_depth(self) -> int:
        """Get the books navigation depth.

        Returns:
            int: the navigation depth.
        """
        return self._navigation_depth

    @property
    def metadata(self) -> List[MetaData]:
        """Get the books metadata.

        Returns:
            List[MetaData]: the books metadata.
        """
        return self._metadata

    @property
    def smils(self) -> List[Smil]:
        """Get the books SMILs.

        Returns:
            List[Smil]: the books SMILs.
        """
        return self._smils

    @property
    def toc_entries(self) -> List[TocEntry]:
        """Return all TOC entries as a list.

        Returns:
            List[TocEntry]: the TOC entries,
        """
        return self._toc_entries

    @property
    def dc_metadata(self) -> List[MetaData]:
        """Get the Dublin Core metadata.

        Returns:
            List[MetaData]: the Dublin Core metadata.
        """
        return [_ for _ in self._metadata if _.name.startswith("dc:") and _.content is not None]

    @property
    def ncc_metadata(self) -> List[MetaData]:
        """Get the Navigation Control Center metadata.

        Returns:
            List[MetaData]: the NCC metadata.
        """
        return [_ for _ in self._metadata if _.name.startswith("ncc:") and _.content is not None]

    @property
    def other_metadata(self) -> List[MetaData]:
        """Get the other metadata (not DC nor NCC).

        Returns:
            List[MetaData]: the other metadata.
        """
        return [_ for _ in self._metadata if not _.name.startswith("dc:") and not _.name.startswith("ncc:") and _.content is not None]

    @property
    def langage(self) -> str:
        """Get the book langage.

        Returns:
            str: the langage.
        """
        metadata = self.get_metadata("dc:language")
        return metadata.content if metadata else ""

    @property
    def charset(self) -> str:
        """Get the character set used in the book.

        Returns:
            str: the charset.
        """
        metadata = self.get_metadata("ncc:charset")
        return metadata.content if metadata else ""

    def get_metadata(self, name: str) -> Union[MetaData | None]:
        """Get metadat by name.

        Note:
        - The full name must be specified, like 'dc:langage' (not simply 'langage').

        Args:
            name (str): the full name (prefix:name).

        Returns:
            Union[MetaData | None]: the MetaData item (or None).
        """
        try:
            return [_ for _ in self._metadata if _.name == name][0]
        except IndexError:
            logger.debug(f"Metadata with name '{name}' not found in the metadata list.")
            return None

    def _populate_entries(self, ncc_document: Document):
        """Process and store the NCC entries (h1 ... h6 tags)."""
        body = ncc_document.get_elements_by_tag_name("body").first()
        for element in body.get_children_by_tag_name().all():
            element_name = element.name
            if element_name in ["h1", "h2", "h3", "h4", "h5", "h6"]:
                level = int(element_name[1])
                id = element.get_attr("id")
                a = element.get_children_by_tag_name("a").first()
                smil_reference = Reference.create_href_or_src(a.get_attr("href"))
                self._toc_entries.append(TocEntry(self.source, id, level, smil_reference, a.text))
        logger.debug(f"Size of toc_entries : {len(self._toc_entries)}.")

    def _populate_metadata(self, ncc_document: Document) -> None:
        """Process and store all metadata."""
        for element in ncc_document.get_elements_by_tag_name("meta").all():
            name = element.get_attr("name")
            if name is not None:
                self._metadata.append(MetaData(name, element.get_attr("content"), element.get_attr("scheme")))
        logger.debug(f"Size of metadata : {len(self._metadata)}.")



from .audio import Audio
from .metadata import MetaData
from .reference import Reference
from .section import Section
from .smil import Smil
from .toc_entry import TocEntry

__all__ = ["Audio", "MetaData", "Reference", "Section", "Smil", "TocEntry"]



from dataclasses import dataclass
from io import BytesIO
import tempfile

from sources.source import DtbSource


@dataclass
class Audio:
    """
    Representation of a <audio/> section in a SMIL file.
    - Defines an audio clip.
    """

    source: DtbSource
    id: str
    src: str
    begin: float
    end: float

    @property
    def duration(self) -> float:
        """Get the duration of a clip, in seconds."""
        return self.end - self.begin

    def get_sound(self, as_bytes_io: bool = False) -> bytes:
        """Get the actual sound data (.wav, .mp3, ...)"""
        return BytesIO(self.source.get(self.src)) if as_bytes_io is True else self.source.get(self.src)



from typing import Any, Callable, List, Union

from loguru import logger


class BaseNavigator:
    """
    This class implements basic metods to navigate in a list of items.

    Navigation means that the class implements a "cursor" that retains the currently selected item of the list.

    Methods:
        - first() : returns the first item.
        - next() : returns the next item or None if there is no next element in the list.
        - prev() : returns the previous item or None if there is no previous element in the list.
        - last() : returns the last item.
        - current() : returns the current item.
        - navigate_to(id) : returns the item by its id.
        - all() : return all items.
        - add_callback(func(item)) : add a callback function, triggered on navigation (when the current item changes).

    Notes:
        - on instanciation, the first item of the list is pointed.
        - the navigate_to(id) method works only if the items have an 'id' attribute.
          If the method fails, no exception is raised, but it simply returns None.
    """

    def __init__(self, items: List[Any], callback: Callable[[Any], None] = None) -> None:
        """Instanciate a `BasicNavigator` class.

        Args:
            items (List[Any]]): a list of elements.
            callback (callback: Callable[[Any], None], optional) : a function to be called on navigation events.

        Raises:
            ValueError: if the supplied list is not iterable
            ValueError: if the list is empty
            ValueError: if all list items ar not of the same type
        """

        # Check if we have a list
        if not isinstance(items, List):
            error_message = "The supplied argument must be a List."
            logger.error(error_message)
            raise ValueError(error_message)

        # Zero length lists are meaningless !
        if len(items) == 0:
            error_message = "An empty list has been supplied."
            logger.error(error_message)
            raise ValueError(error_message)

        # Make sure that all list items are of same kind
        # The relevant type is taken frm the first element in the list
        items_type = type(items[0])
        for item in items:
            if not isinstance(item, items_type):
                error_message = f"All list items must be of same type (in this case {items_type})."
                logger.error(error_message)
                raise ValueError(error_message)

        # Internal attriutes
        self._items: List[Any] = items
        self._id_list: List[str] = None
        self._current_index: int = 0
        self._max_index: int = len(self._items) - 1
        self._on_navigate: Callable[[Any], None] = callback

        logger.debug(f"{type(self)} instance created with {len(self._items)} element(s) of type {items_type}.")

        # Populate the list of ids if the attribute exists
        is_id_attribute_present = "id" in dict(items[0]).keys() if hasattr(items_type, "keys") else hasattr(items[0], "id")
        if is_id_attribute_present:
            if hasattr(items_type, "keys"):
                self._id_list = [_["id"] for _ in self._items]
            else:
                self._id_list = [getattr(_, "id") for _ in self._items]

    @property
    def length(self) -> int:
        return len(self._items)

    def set_callback(self, callback: Callable[[Any], None]) -> None:
        """Sets a navigation callback function.

        It allows to attach an external navigation handler.

        Args:
            func (Callable[[Any],Any]): a callback function which can handle the current item as a parameter.
        """
        self._on_navigate = callback

    def on_first(self) -> bool:
        """Test if the current item is the first one.

        Returns:
            bool: True if the current item is the first one, False otherwise.
        """
        return self._current_index == 0

    def on_last(self) -> bool:
        """Test if the current item is the last one.

        Returns:
            bool: True if the current item is the last one, False otherwise.
        """
        return self._current_index == self._max_index

    def all(self) -> List[Any]:
        """Return all items as a list.

        Returns:
            List[Any]: the navigator items.
        """
        return self._items

    def first(self) -> Any:
        """"""
        """Go to the first item.

        Returns:
            Any: the first item in the list.
        """
        self._current_index = 0
        item = self._items[self._current_index]

        # Perform a callback if required
        if self._on_navigate is not None:
            self._on_navigate(item)

        return item

    def next(self) -> Union[Any, None]:
        """Go to the next item.

        Returns:
            Union[Any, None]: the next item in the list or None if no next item.
        """
        if self._current_index + 1 <= self._max_index:
            self._current_index += 1
            item = self._items[self._current_index]
        else:
            return None

        # Perform a callback if required
        if self._on_navigate is not None:
            self._on_navigate(item)

        return item

    def prev(self) -> Union[Any, None]:
        """Go to the previous item.

        Returns:
            Union[Any, None]: the previous item in the list or None if no previous item.
        """
        if self._current_index - 1 >= 0:
            self._current_index = self._current_index - 1
            item = self._items[self._current_index]
        else:
            return None

        # Perform a callback if required
        if self._on_navigate:
            self._on_navigate(item)

        return item

    def last(self) -> Any:
        """Go to the last item.

        Returns:
            Any: the previous item in the list.
        """
        self._current_index = self._max_index
        item = self._items[self._current_index]

        # Perform a callback if required
        if self._on_navigate:
            self._on_navigate(item)

        return item

    def current(self) -> Any:
        """Get the current item.

        Returns:
            Any: the current item.
        """
        item = self._items[self._current_index]

        return item

    def navigate_to(self, item_id: str | int) -> Union[Any, None]:
        """Navigate to a specific item based on its id.

        Note :
            - If the item has no 'id' attribute, the method does nothing and returns None.

        Args:
            item_id (str): the searched item id

        Returns:
            Union[Any, None]: the targeted item item or None.
        """
        # Can we search by id ?
        if self._id_list is None:
            logger.debug("There is no id attribute present in the list items")
            return None

        try:
            item = self._items[self._id_list.index(item_id)]
            logger.debug(f"Item with id {item_id} of type {type(item)} found.")
            if self._on_navigate is not None:
                self._on_navigate(item)
            return item
        except ValueError:
            logger.debug(f"Item with id {item_id} not found.")
            return None



from dataclasses import dataclass, field
from typing import Tuple

from book.daisybook import DaisyBook
from models import Audio, Section, TocEntry
from navigators.clip_navigator import ClipNavigator
from navigators.section_navigator import SectionNavigator
from navigators.toc_navigator import TocNavigator


class BookNavigatorException(Exception):
    def __init__(self, message):
        super().__init__(message)


@dataclass
class BookNavigator:
    """This class provides book navigation features.

    It provides 3 navigators:
        - A Table Of Content navigator to navigate in the TOC items.
        - A section navigator to navigate in the sections of the current TOC entry.
        - A clip navigator to navigate in the current sections clips.

    Notes:
        - On class instanciation, section and clip pointers are set to the first section and clip respectively.
        - On TOC navigation, section and clip pointers are set to the first section and clip respectively
        - On Section navigation, the clip pointer is set to the first clip of the section

    Raises:
        BookNavigatorException: raised when the supplied instance creation is not a valid `DaisyBook` instance.
    """

    book: DaisyBook
    toc: TocNavigator = field(init=False, default=None)
    sections: SectionNavigator = field(init=False, default=None)
    clips: ClipNavigator = field(init=False, default=None)

    # Private attributes
    _current_entry: TocEntry = field(init=False, default=None)
    _current_section: Section = field(init=False, default=None)
    _current_clip: Audio = field(init=False, default=None)

    @property
    def context(self) -> Tuple[TocEntry, Section, Audio]:
        """Get the current context.

        The context is a tuple containing:
            - the current TOC entry
            - the current section
            - the current clip

        Returns:
            Tuple[TocEntry, Section, Audio]: the current context.
        """
        return (self._current_entry, self._current_section, self._current_clip)

    @property
    def section_text(self) -> str:
        """Get the current section's text.

        Returns:
            str: the current section's text.
        """
        return self._current_section.text.content

    @property
    def toc_text(self) -> str:
        """Get the current TOC entry's text.

        Returns:
            str: the current TOC entry's text.
        """
        return self._current_entry.text

    @property
    def current_toc_entry(self) -> TocEntry:
        """Get the current TOC entry.

        Returns:
            TocEntry: the current TOC entry.
        """
        return self._current_entry

    @property
    def current_section(self) -> Section:
        """Gets the current section.

        Returns:
            Section: the current section.
        """
        return self._current_section

    @property
    def current_clip(self) -> Audio:
        """Gets the current clip.

        Returns:
            Audio: Gets the current clip.
        """
        return self._current_clip

    def __post_init__(self):
        """`BookNavigator` instance post-initialization.

        - Sets the private attributes of the class.
        - Adds callback method to maintain synchronization between the 3 navigators (toc, sections, clips).

        Raises:
            BookNavigatorException: _description_
        """
        if not isinstance(self.book, DaisyBook):
            raise BookNavigatorException("The supplied parameter is not valid.")

        self.toc = TocNavigator(self.book.toc_entries, self.book.navigation_depth)
        self.toc.set_callback(self.on_toc_navigation)
        self._current_entry = self.toc.first()

    def on_toc_navigation(self, toc_entry: TocEntry) -> None:
        self._current_entry = toc_entry
        self.sections = SectionNavigator(toc_entry.sections, self.on_section_navigation)
        self._current_section = self.sections.first()

    def on_section_navigation(self, section: Section) -> None:
        self._current_section = section
        self.clips = ClipNavigator(section._clips, self.on_clip_navigation)
        self._current_clip = self.clips.first()

    def on_clip_navigation(self, clip: Section) -> None:
        self._current_clip = clip



from abc import ABC, abstractmethod
from typing import Any, Union

from loguru import logger

from cache.cache import Cache
from utilities.domlib import Document, DomFactory


class DtbSource(ABC):
    def __init__(self, base_path: str, initial_cache_size=0) -> None:
        """Creates a new `DtbSource`.

        Args:
            base_path (str): a filesystem folder or a web site
            initial_cache_size (int, optional): the size of the resource cache. Defaults to 0.

        Raises:
            ValueError: if the requested cache size is less than 0.
        """
        if initial_cache_size < 0:
            raise ValueError("The cache size cannot be negative.")

        self._base_path = base_path
        self._cache = Cache(max_size=initial_cache_size)

    @property
    def base_path(self) -> str:
        return self._base_path

    @property
    def cache_size(self) -> int:
        return self._cache.maxlen

    @cache_size.setter
    def cache_size(self, size: int) -> None:
        """Resize the resource cache.

        Args:
            new_size (int): the new size.
        """
        self._cache.resize(size)

    @abstractmethod
    def get(self, resource_name: str) -> Union[bytes, str, Document, None]:
        """Get data and return it as a byte array or a string, or None in case of an error.

        When the resource is buffered
            - the method gets it from the buffer
            - if not found in the buffer, it is added to it

        Args:
            resource_name (str): the resource to get (typically a file name)

        Returns:
            bytes | str | None: returned data (str or bytes or None if the resource was not found)
        """
        raise NotImplementedError

    @staticmethod
    def convert_to_document(data: bytes) -> Union[Document | bytes]:
        """Try a conversion of the data to a Document.

        Args:
            data (bytes): the data bytes.

        Returns:
            Union[Document | bytes]: a document or the original bytes.
        """
        # Try to create a Document
        doc = DomFactory.create_document_from_bytes(data)
        if type(doc) is not type(data):
            logger.debug(f"Converted {type(data)} to {type(doc)}.")
        else:
            logger.debug("No conversion happened.")

        return doc

    def do_cache(self, key: str, data: Any) -> None:
        """Store the data into the cache.

        If the cache size is 0, nothing is done.

        Args:
            key (str): the key.
            data (Any): the data to cache.
        """
        self._cache.add(key, data)

    def enable_stats(self, value: bool) -> None:
        self._cache.enable_stats(value)



from .domlib import Document, DomFactory, Element, ElementList
from .fetcher import Fetcher
from .logconfig import LogLevel

__all__ = ["Document", "DomFactory", "Element", "ElementList", "Fetcher", "LogLevel"]



from dataclasses import dataclass, field
from typing import List

from models.audio import Audio
from models.text import Text
from sources.source import DtbSource


@dataclass
class Section:
    """
    Representation of a <par/> section in as SMIL file.
    Objects inside the <par> element will be played at the same time (in parallel).
    """

    source: DtbSource
    id: str
    text: Text

    # Private attributes
    _clips: List[Audio] = field(init=False, default_factory=list)

    @property
    def clips(self) -> List[Audio]:
        return self._clips



from io import BytesIO
from typing import Union
import zipfile

from loguru import logger
from utilities.domlib import Document
from utilities.fetcher import Fetcher
from sources.source import DtbSource


class ZipDtbSource(DtbSource):
    """This class gets data from a ZIP archive (from the filesystem or a web location)."""

    def __init__(self, base_path) -> None:
        super().__init__(base_path, 0)
        self.bytes_io: BytesIO = None

        if Fetcher.is_available(base_path) is False:
            raise FileNotFoundError

        # Get the zip data
        self.bytes_io = BytesIO(Fetcher.fetch(base_path))

        # Check if we have a good ZIP file
        if zipfile.is_zipfile(self.bytes_io):
            logger.debug(f"{base_path} is a valid ZIP archive.")
        else:
            raise FileNotFoundError

    def get(self, resource_name: str) -> Union[bytes, Document, None]:
        # Try to get data from the cached resources
        cached_data = self._cache.get(resource_name)
        if cached_data is not None:
            return cached_data

        # Retrieve the resource fron the ZIP file
        with zipfile.ZipFile(self.bytes_io, mode="r") as archive:
            try:
                data = archive.read(resource_name)
            except KeyError:
                logger.error(f"Error: archive {self._base_path} does not contain file {resource_name}")
                return None

        # Try to create a Document
        doc = DtbSource.convert_to_document(data)

        # Eventualy cache the resource
        self.do_cache(resource_name, doc)

        return doc



"""Classes to encapsulate and simplify the usage of the xml.dom.minidom library."""

import re
import urllib.request
import xml.dom.minidom
from dataclasses import InitVar, dataclass, field
from typing import Dict, List, Optional, Union
from urllib.error import HTTPError, URLError
from xml.dom.minidom import parseString as xdm_parse_string
from xml.parsers.expat import ExpatError

from loguru import logger


@dataclass
class Element:
    """Representation of a DOM element."""

    xml_node: InitVar[xml.dom.minidom.Element]  # The node in the xml.dom.minidom referential

    # Internal attributes
    _xml_node: xml.dom.minidom.Element = field(init=False, default=None)
    _children: "ElementList" = field(init=False, default=None)

    def __post_init__(self, xml_node: xml.dom.minidom.Element):
        """Post initialization of the Element instance."""
        if not isinstance(xml_node, xml.dom.minidom.Element):
            return

        self._xml_node = xml_node
        self._children = DomFactory.create_element_list(self._xml_node.childNodes)

    def _get_text(self, root: xml.dom.minidom.Node, _text: str = "") -> str:
        """Get text from the root element and its children.

        Notes:
            - This method is recursive.
            - Do not call directly (private method).

        Args:
            root (xml.dom.minidom.Node): the root element
            _text (str, optional): the current text. Defaults to "".

        Returns:
            str: the full string.
        """
        child: xml.dom.minidom.Element
        for child in root.childNodes:
            match child.nodeType:
                case xml.dom.minidom.Node.TEXT_NODE:
                    _text += child.nodeValue
                case xml.dom.minidom.Node.ELEMENT_NODE:
                    # Recurse here !
                    _text = self._get_text(child, _text)
                case _:
                    # Do nothing !
                    ...

        return _text

    @property
    def is_void(self) -> bool:
        """Test if the class is fully instanciated.

        Returns:
            bool: True if the class is correctly intanciated, False otherwise.
        """
        return isinstance(self._xml_node, xml.dom.minidom.Element) is False

    @property
    def has_children(self) -> bool:
        return self._children.size > 0

    @property
    def name(self) -> Union[str, None]:
        """Get the element's name (tag name)."""
        if self.is_void:
            return None

        return self._xml_node.tagName if not self.is_void else None

    @property
    def value(self) -> Union[str, None]:
        if self.is_void:
            return None

        return self._xml_node.firstChild.nodeValue if self._xml_node.hasChildNodes() else None

    @property
    def text(self) -> Union[str, None]:
        """Returns a string with no carriage returns and duplicate spaces."""
        if self.is_void:
            return None

        text = self._get_text(self._xml_node)
        return re.sub(r"\s+", " ", text).strip() if len(text) else None

    @property
    def parent(self) -> Union["Element", None]:
        """Get the parent element.

        Returns:
            Element: an element or None
        """
        if self.is_void:
            return None

        return Element(self._xml_node.parentNode) if self._xml_node.parentNode else None

    def get_attr(self, attr: str) -> Union[str, None]:
        """Get the value of an attribute."""
        if self.is_void:
            return None

        attr = self._xml_node.getAttribute(attr)
        return attr if attr else None

    def get_children_by_tag_name(self, tag_name: Optional[str] = "") -> Union["ElementList", None]:
        """Get all child elements by tag name (or all if no tag_name is specified).

        Args:
            tag_name (str, optional): the searched tag name. Defaults to "".

        Returns:
            ElementList: an element list or None.
        """
        if self.is_void or isinstance(tag_name, str) is False:
            return None

        if tag_name.strip() == "":
            return self._children

        result = ElementList()
        child: Element
        for child in self._children.all():
            if child.name == tag_name:
                result.elements.append(child)

        return result


@dataclass
class ElementList:
    elements: List[Element] = field(default_factory=list)

    @property
    def size(self) -> int:
        """Get the number of elements."""
        return len(self.elements)

    def append(self, element: Element) -> None:
        """Add an element."""
        if element and isinstance(element, Element):
            self.elements.append(element)

    def first(self) -> Element | None:
        """Get the first element."""
        return self.elements[0] if self.size > 0 else None

    def all(self) -> List[Element]:
        """Get all elements"""
        return self.elements


@dataclass
class Document:
    """This class represents a whole xml or html document."""

    xml_node: InitVar[xml.dom.minidom.Document]

    # Internal attributes
    _xml_node: xml.dom.minidom.Element = field(init=False, default=None)

    def __post_init__(self, xml_node: xml.dom.minidom.Document):
        """Post initialization of the Document instance."""
        if not isinstance(xml_node, xml.dom.minidom.Document):
            return
        self._xml_node = xml_node

    def get_element_by_id(self, id: str) -> Union[Element, None]:
        """Get an element by its id"""
        for elt in self._xml_node.getElementsByTagName("*"):
            if elt.getAttribute("id") == id:
                return Element(xml_node=elt)
        return None

    def get_elements_by_tag_name(self, tag_name: str, filter: Dict = {}, having_parent_tag_name: str = None) -> ElementList:
        """
        Get elements by tag name.

        A filter on attributes can be specified. The form is `{"attribute_name": "attribute_value"}`.

        Since xml.minidom does a recursive search, a parent tag name can be specified to filter out unwanted elements.

        Args:
            tag_name (str): the seaerched tag name
            filter (Dict, optional): the attribute filter. Defaults to {}.
            having_parent_tag_name (str, optional): the parent tag name filter. Defaults to None.

        Returns:
            ElementList | None: the searched element (may be empty).
        """
        if self._xml_node is None:
            return ElementList()

        logger.debug(f"tag_name: {tag_name}, filter: {filter}, parent_tag_name: {having_parent_tag_name}")
        xdm_nodes = self._xml_node.getElementsByTagName(tag_name)

        # Filter data
        if having_parent_tag_name:
            xdm_node_list = []
            for xdm_element in xdm_nodes:
                if xdm_element.parentNode.tagName == having_parent_tag_name:
                    xdm_node_list.append(xdm_element)
        else:
            xdm_node_list = xdm_nodes

        # No filter : get all elements
        if len(filter.items()) == 0:
            return DomFactory.create_element_list(xdm_node_list)

        # With filtering
        result = ElementList()
        for elt in xdm_node_list:
            for k, v in filter.items():
                attr = elt.getAttribute(k)
                if attr == v:
                    result.append(Element(xml_node=elt))

        return result


class DomFactory:
    """This class holds a collection of static methods to create various class instances."""

    @staticmethod
    def create_document_from_string(string: str) -> Union[Document, None]:
        """Create a Document from a string.
        Args:
            string (str): The string to parse

        Returns:
            Document | None: a Document or None
        """
        # Type chack
        if not isinstance(string, str) or len(string) == 0:
            return None

        try:
            xdm_document = xdm_parse_string(string)
        except ExpatError as e:
            logger.error(f"An xml.minidom parsing error occurred. The code is {e.code}.")
            return None
        return Document(xml_node=xdm_document)

    @staticmethod
    def create_document_from_url(url: str) -> Union[Document, None]:
        """Create a Document from an URL.
        Args:
            url (str): The URL to parse

        Returns:
            Document | None: a Document or None
        """
        try:
            response = urllib.request.urlopen(url)
            data = response.read()
            return Document(xml_node=xdm_parse_string(data))
        except HTTPError as e:
            logger.error(f"HTTP error: {e.code} {e.reason} ({url})")
        except URLError as e:
            logger.error(f"URL error: {e.reason} ({url})")
        return None

    @staticmethod
    def create_document_from_bytes(data: bytes) -> Union[Document, bytes]:
        if not isinstance(data, bytes):
            return data

        try:
            string = data.decode("utf-8")
            return DomFactory.create_document_from_string(string)
        except UnicodeDecodeError:
            ...

        return data

    @staticmethod
    def create_element_list(xml_nodes: List[xml.dom.minidom.Element]) -> ElementList:
        """Create an Element list from a list of xml.minidom nodes.

        Args:
            nodes (List[xml.dom.minidom.Element]): an xml.minidom element list.

        Returns:
            ElementList: a list of Element
        """
        result = ElementList()
        for node in xml_nodes:
            if isinstance(node, xml.dom.minidom.Element):
                result.elements.append(Element(xml_node=node))

        return result



from dataclasses import dataclass, field
from typing import List, override

from loguru import logger

from models import TocEntry
from navigators.base_navigator import BaseNavigator


@dataclass
class TocNavigator(BaseNavigator):
    """
    This class provides method to navigate in table of contents of a digital talking book.
    The TOC is provided by the ncc.html file, common in Daisy 2.02 projects.

    Notes :
        - It overrides the methods of its `Navigator` base class.
        - It also provides methods to generate a TOC of the book
    """

    toc_entries: List[TocEntry]
    navigation_depth: int

    # Internal attributes
    _max_nav_level: int = field(init=False, default=0)
    _current_nav_level: int = field(init=False, default=0)

    def __post_init__(self):
        """Postinitialitation of the dataclass.

        Action(s):
            - Initialize the base class
            - Set the max. navigation level
        """
        super().__init__(self.toc_entries)
        self._max_nav_level = self.navigation_depth
        logger.debug(f"Initialization of class {type(self)} done. Max. naigation level is {self._max_nav_level}.")

    @property
    def filter_is_active(self) -> bool:
        return self._current_nav_level != 0

    def set_nav_level(self, level: int) -> int:
        """Set the navigation level.

        Args:
            level (int): the requested navigation level

        Returns:
            int: the actual navigation level
        """
        # Check
        if level < 0 or level > self._max_nav_level:
            return self._current_nav_level

        self._current_nav_level = level
        return self._current_nav_level

    def get_nav_level(self) -> int:
        """Get the current navigation level.

        Returns:
            int: the current navigation level.
        """
        return self._current_nav_level

    def increase_nav_level(self) -> int:
        """Increase the navigation level.

        Returns:
            int: the updated navigation level
        """
        return self.set_nav_level(self._current_nav_level + 1)

    def decrease_nav_level(self) -> int:
        """Decrease the navigation level.

        Returns:
            int: the updated navigation level
        """
        return self.set_nav_level(self._current_nav_level - 1)

    def reset_nav_level(self) -> int:
        """Reset (remove) the navigation level.

        Returns:
            int: the updated navigation level
        """
        return self.set_nav_level(0)

    @override
    def first(self) -> TocEntry:
        """Get the first NCC entry.
        - If a level filter is active, it is taken into account.

        Returns:
            NccEntry: the first entry
        """
        item: TocEntry = super().first()

        if self.filter_is_active:
            # Enumerate upwards
            while item is not None:
                if item.level == self._current_nav_level:
                    break
                item = super().next()

        return item

    @override
    def last(self) -> TocEntry:
        """Get the last NCC entry.
        - If a level filter is active, it is taken into account.

        Returns:
            NccEntry: the last entry
        """
        item: TocEntry = super().last()
        if self.filter_is_active:
            while item is not None:
                if item.level == self._current_nav_level:
                    break
                item = super().prev()

        return item

    @override
    def next(self) -> TocEntry | None:
        """Get the next NCC entry.
        - If a level filter is active, it is taken into account.

        Returns:
            NccEntry: the next entry
        """
        item: TocEntry = super().next()
        if self.filter_is_active:
            while item is not None:
                if item.level == self._current_nav_level:
                    break
                item = super().next()

        return item

    @override
    def prev(self) -> TocEntry | None:
        """Get the previous NCC entry.
        - If a level filter is active, it is taken into account.

        Returns:
            NccEntry: the previous entry
        """
        item: TocEntry = super().prev()
        if self.filter_is_active:
            while item is not None:
                if item.level == self._current_nav_level:
                    break
                item = super().prev()

        return item

    def generate_toc(self, format: str) -> str:
        """Generate a TOC of the current book.

        Supported formats:
            - `md-list`    : a Markdown list (*)
            - `md-headers` : Markdown headers (#)
            - `html-headers` : HTML headers (<h1/> to <h6/>)

        Args:
            format (str): the requested format

        Raises:
            ValueError: raised when a format is not handled.

        Returns:
            str: the formatted TOC
        """
        result = ""
        if isinstance(format, str) is False:
            return result

        match format.lower():
            case "md-list":
                for entry in self.toc_entries:
                    result += f'{"   " * (entry.level-1)}* {entry.text}\n'
            case "md-headers":
                for entry in self.toc_entries:
                    result += f'{"#" * (entry.level):6} {entry.text}\n'
            case "html-headers":
                for entry in self.toc_entries:
                    result += f"<h{(entry.level)}>{entry.text}</h{(entry.level)}>\n"
            case _:
                raise ValueError(f"Invalid format ({format}).")

        return result



"""Resource cacheing classes"""

from collections import deque
from dataclasses import InitVar, dataclass, field
from typing import Any

from loguru import logger

from cache.cachestats import CacheStats


@dataclass
class _CacheItem:
    """This class represents a cached resource.

    Note:
    - It is intended for internal use.
    """

    key: str
    data: Any

    def __post_init__(self):
        """Class post initilization."""
        logger.debug(f"The cache item '{self.key}' has been created. Its type is {type(self.data)}.")

    @property
    def type(self) -> type:
        """Get the cached data type.

        Returns:
            type: the data type
        """
        return type(self.data)


@dataclass
class Cache:
    """Representation of resource cache"""

    max_size: InitVar[int] = 0
    with_stats: InitVar[bool] = False

    # Internal attributes
    _items: deque[_CacheItem] = field(init=False, default_factory=deque)
    _with_stats: bool = field(init=False, default=False)
    _stats: CacheStats = field(init=False, default_factory=CacheStats)

    def __post_init__(self, max_size: int, with_stats: bool) -> None:
        """Cache post initialize.

        Args:
            max_size (int): the cache size.
        """

        if max_size < 0:
            logger.warning(f"The cache size must be positive. {max_size} was supplied: cache size set to 0.")
            max_size = 0

        self._items = deque(maxlen=max_size)
        self._with_stats = with_stats
        logger.debug(f"Cache created. Size: {max_size}. Statistics collection is {'active' if self._with_stats else 'inactive'}.")

    @property
    def maxlen(self) -> int:
        return self._items.maxlen

    def get_stats(self) -> dict:
        """Get the cache statistics.

        Returns:
            dict: a dict with the statistics.
        """
        return self._stats.get_stats()

    def enable_stats(self, value: bool) -> None:
        """Enable or disable statistics collection.

        Args:
            value (bool): True -> enable, False -> disable.
        """
        self._with_stats = value
        logger.debug(f"Cache statistics collection is {'active' if self._with_stats else 'inactive'}.")

    def resize(self, new_size: int) -> None:
        """Resize the cache.

        Args:
            new_size (int): the new size
        """
        # Checks
        if not isinstance(new_size, int) or (new_size < 0) or (new_size == self._items.maxlen):
            return

        # Cache migration
        logger.debug(f"Resizing the cache from {self._items.maxlen} to {new_size}.")
        new_cache = deque(maxlen=new_size)
        self._items.reverse()
        if new_cache.maxlen > 0:
            for index, item in enumerate(self._items):
                if index > new_cache.maxlen - 1:
                    break
                new_cache.appendleft(item)

        self._items = new_cache
        logger.debug(f"The cache size now is {self._items.maxlen}.")

    def add(self, key: str, data: Any) -> None:
        """Add data into the cache.

        Notes :
            - If the cache max. length is 0, nothing is done.
            - If the kex exists in the cache, data is updated.
            - If the addition would overfill the cache, the oldest item is removed before adding the new one.

        Args:
            key (str): the key.
            data (Any): the data.
        """

        # Checks
        if self._items.maxlen == 0:
            return

        try:
            # Check if item exists and update the current data
            item_index = [_.key for _ in self._items].index(key)
            self._items[item_index].data = data
            logger.debug(f"Resource '{key}' in the cache (index={item_index}) has been updated.")
            return
        except ValueError:
            # Otherwise append the item
            self._items.append(_CacheItem(key, data))
            logger.debug(f"Item '{key}' added into the cache as {type(data)}.")

    def get(self, key: str) -> Any | None:
        """Get data from the cache.

        Args:
            resource_name (str): the requested resource

        Returns:
            Any | None: the found data or None
        """
        # No cache, no data
        if self._items.maxlen == 0:
            logger.debug("There is no cache size defined. Returning 'None'.")
            return None

        try:
            # Try to find the key
            item_index = [_.key for _ in self._items].index(key)
            logger.debug(f"Item '{key}' found in the cache (index={item_index}).")
            if self._with_stats:
                self._stats.hit(key)
            return self._items[item_index].data
        except ValueError:
            # Key not found
            logger.debug(f"Item '{key}' not found in the cache.")
            if self._with_stats:
                self._stats.miss(key)
            return None



from dataclasses import dataclass, field

from loguru import logger

from utilities.domlib import Document

from models.reference import Reference
from sources.source import DtbSource


@dataclass
class Text:
    """Representation of a text fragment in a text source file."""

    source: DtbSource
    id: str
    reference: Reference

    # Internal attributes
    _content: str = field(init=False, default=None)

    @property
    def content(self) -> str:
        """Get the text fragment identified by the reference.

        Note:
        - To avoid multiple resource accesses, the text is loaded from the source only once.

        Returns:
            str: the text fragment
        """
        if self._content is None:
            self._parse()
        return self._content

    def _parse(self) -> None:
        """Get the text from a resource.

        Notes:
        - If the text has already been retrieved, return the instances text content.


        Returns:
            str: the text.
        """
        # Check if text already here
        if self._content is not None:
            logger.debug(f"Content {self.reference.resource}/{self.reference.fragment} is already present.")
            return self._content

        # Get it from the source
        logger.debug(f"Loading text from {self.reference.resource}, fragment id is {self.reference.fragment}.")
        data = self.source.get(self.reference.resource)

        # The fetched data must be a Document
        if isinstance(data, Document) is False:
            logger.error(f"The retrieval attempt of {self.reference.resource} as Document failed.")
            self._content = ""
            return

        # Find the text identified by its id
        element = data.get_element_by_id(self.reference.fragment)
        if element is not None:
            self._content = element.text
            logger.debug(f"Text with id {self.reference.fragment} found in {self.reference.resource}.")
            return
        else:
            logger.error(f"Could not retrieve element {self.reference.fragment} in the {self.reference.resource} Document.")

        self._content = ""



"""This is the package file."""

from .book import DaisyBook, DaisyBookException
from .cache import Cache, CacheStats
from .models import Audio, MetaData, Reference, Section, Smil
from .navigators import BaseNavigator, BookNavigator, BookNavigatorException, ClipNavigator, SectionNavigator
from .sources import DtbSource, FolderDtbSource, ZipDtbSource
from .utilities import Document, DomFactory, Element, ElementList, Fetcher, LogLevel

__all__ = [
    "DaisyBook",
    "DaisyBookException",
    "Cache",
    "CacheStats",
    "Audio",
    "MetaData",
    "Reference",
    "Section",
    "Smil",
    "BaseNavigator",
    "BookNavigator",
    "BookNavigatorException",
    "ClipNavigator",
    "SectionNavigator",
    "DtbSource",
    "FolderDtbSource",
    "ZipDtbSource",
    "Document",
    "DomFactory",
    "Element",
    "ElementList",
    "Fetcher",
    "LogLevel",
]



from dataclasses import dataclass, field
from typing import List

from loguru import logger

from models.reference import Reference
from models.section import Section
from models.smil import Smil
from sources.source import DtbSource


@dataclass
class TocEntry:
    """Representation of an entry in the NCC file."""

    source: DtbSource
    id: str
    level: int
    smil_reference: Reference
    text: str

    # Internal attributes
    _smil: Smil = field(init=False, default=None)

    def __post_init__(self):
        # Build the SMIL from its reference
        self._smil = Smil(self.source, self.smil_reference)
        logger.debug(f"Smil set from {self.smil_reference}")

    @property
    def smil(self) -> "Smil":
        """Get the attached SMIL.

        Returns:
            Smil: the SMIL.
        """
        return self._smil

    @property
    def sections(self) -> List[Section]:
        return self._smil.sections



"""Resources operations"""

from dataclasses import dataclass, field
from http.client import HTTPResponse
from pathlib import Path
from urllib.error import HTTPError, URLError
import urllib.request
from loguru import logger
import urllib


@dataclass
class Fetcher:
    """This class provides static methods to
    - check resource availability
    - fetch resources

    It automatically handles the location of the resource (file system or web).
    """

    fetched_bytes: int = field(init=False, default=0)
    access_count: int = field(init=False, default=0)

    @staticmethod
    def get_stats() -> dict:
        """Get the fetcher statistics.

        Returns:
            dict: a dict with following keys:
            - "access_count" : the number of times the fetcher was used.
            - "fetched_bytes" : the number of retrieved bytes.
        """
        return {
            "access_count": Fetcher.access_count,
            "fetched_bytes": Fetcher.fetched_bytes,
        }

    @staticmethod
    def is_on_web(resource_path: str) -> bool:
        """Test if the resource is located on the web.

        Args:
            resource (str): the resource (full path).

        Returns:
            bool: True if there is '://' in the full resource path, False otherwise.
        """
        return "://" in resource_path

    @staticmethod
    def is_available(resource_path: str) -> bool:
        """Checks the availability of a resource.

        Args:
            resource (str): the resource (folder or web location).

        Returns:
            bool: True if the resource is found, False otherwise.
        """
        logger.debug(f"Checking availability of '{resource_path}'.")

        # Check
        if not isinstance(resource_path, str):
            logger.debug("No valid data supplied.")
            return False

        Fetcher.access_count += 1
        if Fetcher.is_on_web(resource_path):
            # Check web availability
            try:
                response = urllib.request.urlopen(resource_path)
                if isinstance(response, HTTPResponse) and response.getcode() == 200:
                    logger.debug("Web check success. Error code is 200.")
                    return True
                else:
                    logger.debug("Web check failed. Response is not of type HTTPResponse or error code is not 200.")
                    return False
            except HTTPError as e:
                error_code = e.getcode()
                if error_code in (200, 403):  # Code 403 is not necessarily an error !
                    logger.debug(f"Web check success. Error code is {error_code}. Codes 200 and 403 are OK.")
                    return True
                logger.debug(f"Web check fails. Error code is {error_code}.")
                return False
            except URLError:
                logger.debug(f"Web check fails wit an URL error. The failing URL is {resource_path}.")
                return False
        else:
            # Check file system availability
            if Path(resource_path).exists():
                logger.debug("File check success.")
                return True
            else:
                logger.debug(f"File check fails. The Path is {resource_path}")
                return False

    @staticmethod
    def fetch(resource_path: str) -> bytes:
        """Fetch a given resource

        Args:
            resource (str): the resource to fetch (full path).

        Returns:
            bytes: the fetched bytes (or b'').
        """
        Fetcher.access_count += 1

        logger.debug(f"Fetching '{resource_path}'.")
        # Check
        if not isinstance(resource_path, str):
            logger.debug("No valid data supplied.")
            return b""

        if Fetcher.is_on_web(resource_path):
            # Get data from web
            try:
                response = urllib.request.urlopen(resource_path)
                if isinstance(response, HTTPResponse) and response.getcode() == 200:
                    data = response.read()
                    Fetcher.fetched_bytes += len(data)
                    logger.debug(f"Fetched {len(data)} bytes from {resource_path}.")
                    return data
                else:
                    logger.debug(f"Nothing fetched from {resource_path}.")
                    return b""
            except URLError:
                logger.debug(f"URL error: {resource_path}.")
                return b""
        else:
            # Get data from file system
            try:
                with open(resource_path, "rb") as file:
                    data = file.read()
                    Fetcher.fetched_bytes += len(data)
                    logger.debug(f"Fetched {len(data)} bytes from {resource_path}.")
                    return data
            except FileNotFoundError:
                logger.debug(f"Nothing fetched from {resource_path} (not found).")
            except IsADirectoryError:
                logger.debug(f"Nothing fetched from {resource_path} (the resource is a folder).")

            return b""
